#class auto

#define SPI_SER_B
#define SPI_CLK_DIVISOR			5
#define SPI_CLOCK_MODE 1
#define SERB_USEPORTD
#use SPI.LIB


#define MMC_SECTORSZ 512
#define MMC_SECTORL2 9

//data block transfer tokens
#define MMC_START_BLOCK_SREAD		0xFE
#define MMC_START_BLOCK_MREAD		0xFE
#define MMC_START_BLOCK_SWRITE	0xFE
#define MMC_START_BLOCK_MWRITE	0xFC
#define MMC_STOP_TRANS				0xFD

//responses
#define MMC_R1_OK      	0x00
#define MMC_R1_IDLING   0x01

//data responses
#define MMC_ACCEPTED		0x05
#define MMC_CRCERR		0x0B
#define MMC_WRERR			0x0D
// mask
#define MMC_DRMASK		0x1F
//busy signal
#define MMC_BUSY			0x00




// MMC SPI supported command set: name, CMD#
// (not necessarily supported on this implementation)
#define MMC_GO_IDLE_STATE				00
#define MMC_SEND_OP_COND				01
#define MMC_SEND_CSD						09
#define MMC_SEND_CID						10
#define MMC_STOP_TRANSMISSION			12
#define MMC_SEND_STATUS					13
#define MMC_SET_BLOCKLEN				16
#define MMC_READ_SINGLE_BLOCK			17
#define MMC_READ_MULTIPLE_BLOCK		18
#define MMC_WRITE_BLOCK					24
#define MMC_WRITE_MULTIPLE_BLOCK		25
#define MMC_PROGRAM_CSD					27
#define MMC_SET_WRITE_PROT				28
#define MMC_CLR_WRITE_PROT				29
#define MMC_SEND_WRITE_PROT			30
#define MMC_TAG_SECTOR_START			32
#define MMC_TAG_SECTOR_END				33
#define MMC_UNTAG_SECTOR				34
#define MMC_TAG_ERASE_GROUP_START	35
#define MMC_TAG_ERASE_GROUP_END		36
#define MMC_UNTAG_ERASE_GROUP			37
#define MMC_ERASE							38
#define MMC_READ_OCR						58
#define MMC_CRC_ON_OFF					59


// Timeout constants in bytes time
// NCR=max 8 bytes
#define MMC_TMOUT_NCR					8
// NAC=max (10*TAAC*f+100*NSAC)/8
#define MMC_TMOUT_NAC					1000
// wait after reset, technically not a timeout...
#define MMC_TMOUT_RST					10
// tired of getting busy signal (there is no spec about this)
#define MMC_TMOUT_BSY					1000


// Miscellaneous constants
#define DUMMY	0xFF
#define MMC_TIMEOUT     DUMMY
//CRC for CMD0
#define MMC_CMD0CRC 0x4A

#define MMC_TRANSFER_OK 0xAA

const static unsigned char DUMMYBUF[]={DUMMY,DUMMY};


/* Init related hardware
*/
void MMC_hardwareinit(void)
{
	SPIinit();
	BitWrPortI ( PEDDR, &PEDDRShadow, 1, 7 );	// CS = output
	BitWrPortI ( PEFR, &PEFRShadow, 0, 7 );	// claim I/O pin
	BitWrPortI ( PFDDR, &PFDDRShadow, 1, 0 );	// PWR = output
	BitWrPortI ( PFFR, &PFFRShadow, 0, 0 );	// claim I/O pin
	BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
	BitWrPortI ( PFDR, &PFDRShadow, 0, 0 );	// no PWR
};

/* Send a command to the MMC card, crc is 7-bit CRC right aligned
*/
void MMC_SndCmd (unsigned char cmd, unsigned long data, unsigned char crc)
{
unsigned char *a,buf[6];
int i;

	a=(unsigned char *) &data;		// LSB first, pay attention to endianness if porting
	buf[0]=cmd|0x40;					// add transmitter bit for host command, start bit is 0 because of cmd number
	i=4;
	do {
		buf[i]=(*(a++));				// MSB first, pay attention to endianness if porting
   } while(--i);
	buf[5]=((crc<<1)|0x01);			// add end bit
   SPIWrite(buf,6);					// Send
}

/* get a response from the MMC card, and return it
Timeout must be >0, on timeout, last response is DUMMY,
so on timeout, returned value is effectively DUMMY, which is MMC_TIMEOUT
*/
char MMC_GetResponse(int timeout)
{
unsigned char response;

	while(timeout--){
   	SPIRead(&response,1);
		if(response!=DUMMY)
			break;
	}
	return(response);
}


/* Busy waits until card reports init OK, error, or timeout
Returns MMC_R1_OK on success, MMC_R1_IDLING on timeout, error R1 on error
*/
unsigned char MMC_init (unsigned int timeout)
{
unsigned char response;
int t;

	BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
	t=MMC_TMOUT_RST;						// wait some time after hardware reset
   while(t--)
		SPIWrite(DUMMYBUF,1);
	BitWrPortI ( PEDR, &PEDRShadow, 0, 7 );	// enable CS

	MMC_SndCmd(MMC_GO_IDLE_STATE,0L,MMC_CMD0CRC);	// reset card
/* after getting response to MMC_GO_IDLE_STATE, repeatedly poll
card until it responds MMC_R1_OK or we get tired, whatever comes first */
	while((response=MMC_GetResponse(MMC_TMOUT_NAC))==MMC_R1_IDLING){
      if(!timeout--)
      	break;
		BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
		SPIWrite(DUMMYBUF,1);
		BitWrPortI ( PEDR, &PEDRShadow, 0, 7 );	// enable CS
		MMC_SndCmd(MMC_SEND_OP_COND,0L,DUMMY);			// start/poll initialization process
   }
	BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
	SPIWrite(DUMMYBUF,1);			// keep clock going for some time
	return (response);
}

/* Returns after busy signal is removed, or timeout
Returns read value if != MMC_BUSY
*/
unsigned char MMC_bsywait(int timeout)
{
unsigned char response;

	while(timeout--){
		SPIRead(&response,1);
		if(response!=MMC_BUSY)
   	 	break;
  }
  return(response);
}

/* Read a block of data, leave it at buf
Returns MMC_TRANSFER_OK if OK, errored R1 or data token on error
Not very friendly on error, feel free to modify if you really need to know
what is wrong
*/
unsigned char MMC_rdblk(unsigned char *buf,unsigned int len,)
{
unsigned char response,CRCbuf[2];

	if ((response=MMC_GetResponse(MMC_TMOUT_NCR))==MMC_R1_OK) {
		if ((response=MMC_GetResponse(MMC_TMOUT_NAC))==MMC_START_BLOCK_SREAD) {
			SPIRead(buf,len);       // get block
			SPIRead(CRCbuf,2);		// get CRC (ignored)
			response=MMC_TRANSFER_OK;
		}
	}
   return(response);
}


/* Read a sector, leave data at buf
uses MMC_rdblk, so see it
*/
unsigned char MMC_RdSector(unsigned long sector,unsigned char *buf)
{
unsigned char response;
int i;
unsigned long address;

	address=sector << MMC_SECTORL2;
	BitWrPortI ( PEDR, &PEDRShadow, 0, 7 );	// enable CS
	MMC_SndCmd(MMC_READ_SINGLE_BLOCK,address,DUMMY);
   response=MMC_rdblk(buf,MMC_SECTORSZ);
	BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
	SPIWrite(DUMMYBUF,1);
	return(response);
}


/* Write a sector with data in buf
Returns MMC_TRANSFER_OK on success, errored R1 or data token/non busy on error
Not very friendly on error, feel free to modify if you really need to know
what is wrong
 */
unsigned char MMC_WrSector (unsigned long sector, unsigned char *buf)
{

unsigned char response;
int i;
const static unsigned char wrcmd[]={DUMMY,MMC_START_BLOCK_SWRITE};
unsigned long address;

	address=sector << MMC_SECTORL2;
	BitWrPortI ( PEDR, &PEDRShadow, 0, 7 );	// enable CS
	MMC_SndCmd(MMC_WRITE_BLOCK,address,DUMMY);
	if((response=MMC_GetResponse(MMC_TMOUT_NCR))==MMC_R1_OK) {
      SPIWrite(wrcmd,2); 						// send start token
		SPIWrite(buf,MMC_SECTORSZ);
		SPIWrite(DUMMYBUF,2);		         // send dummy CRC bytes
      response=MMC_GetResponse(MMC_TMOUT_NCR)&MMC_DRMASK;
		if(response==MMC_ACCEPTED)
      	if((response=MMC_bsywait(MMC_TMOUT_BSY))==DUMMY)
				response=MMC_TRANSFER_OK;
	}
	BitWrPortI ( PEDR, &PEDRShadow, 1, 7 );	// disable CS
	SPIWrite(DUMMYBUF,1);
	return(response);
}


unsigned char buffer[MMC_SECTORSZ];


int printsector(long sector)
{
unsigned char rc;
int i;

	if((rc=MMC_RdSector(sector, buffer))==MMC_TRANSFER_OK){
	   for(i=0;i<MMC_SECTORSZ;i++){
      	if((i&0xFF0)==i)
   			printf("\n%04X\t",i);
   		printf("%02X ",buffer[i]);
	   }
   }
   else printf("result: %02X\n",rc);
   return(rc);
}


int main( void )
{
unsigned char rc;
int i;
long addr;
char aux[20];

   for(i=0;i<MMC_SECTORSZ;i++)
   	buffer[i]=0;

	MMC_hardwareinit();
	BitWrPortI ( PFDR, &PFDRShadow, 1, 0 );	// PWR
	if((rc=MMC_init(1000))!=MMC_R1_OK){
		printf ("init: %02X\n", rc);
      exit(1);
	}
	if((rc=MMC_RdSector(0, buffer))==MMC_TRANSFER_OK){
        	printf("File system descriptor for first partition: ");
   		switch(buffer[0x1be+4]){
   		case 0:
         	printf("Empty");
            exit(1);
   		case 4:
         	printf("DOS 16-bit FAT <32MB\n");
            break;
   		case 6:
         	printf("DOS 16-bit FAT >32MB\n");
            break;
			default:
         	printf("Unsupported\n");
            exit(1);
         }
          addr=(*((long *)&buffer[0x1be+8]));
			 printf("Address of first sector in first partition: %08lX\n\n",addr);
		 	 if((rc=MMC_RdSector(addr, buffer))==MMC_TRANSFER_OK){
          	buffer[0xb]=0;
            printf("OEM name: %s\n",&buffer[3]);
          	buffer[0x3e]=0;
            printf("Volume label, file system type: %s\n",&buffer[0x2b]);
            addr+=((*((int *)&buffer[0x0e]))+buffer[0x10]*(*((int *)&buffer[0x16])));
            printf("Sectors per cluster: %d\n",buffer[0x0D]);
            printf("Root directory entries: %d\n",*((int *)&buffer[0x11]));
            printf("Root directory sector: %ld\n",addr);
            if(printsector(addr)!=MMC_TRANSFER_OK)
            	exit(1);
            printf("\n\nDirectory:\n\n");
            i=0;
            while((buffer[i]!=0)&&(i<512)){
            	if((buffer[i]!=0xE5)&&(!(buffer[i+11]&0x0C))){
               	strncpy(aux,&buffer[i],8);
                  aux[8]=0;
                  printf("%s\t",aux);
               	strncpy(aux,&buffer[i+8],3);
                  aux[3]=0;
                  printf("%s\t",aux);
                  if(buffer[i+11]&0x10)
							printf("<DIR>\t");
                  else
							printf("%ld\t",*((long *)&buffer[i+28]));
                  printf("%02d/%02d/%02d\t",(int)(buffer[i+24]&0x1F),(*((int *)&buffer[i+24])&0x01E0)>>5,(int)((buffer[i+25]&0xFE)>>1)-20);
                  printf("%02d:%02d:%02d\n",(int)(buffer[i+23]>>3),(*((int *)&buffer[i+22])&0x07E0)>>5,(int)(buffer[i+25]&0x1F));
               }
               i+=32;
            }
		   }
		   else printf("result: %02X\n",rc);
   }
   else printf("result: %02X\n",rc);

	BitWrPortI ( PFDR, &PFDRShadow, 0, 0 );	// no PWR
}


